@@ -8,20 +8,17 @@ import keyword
 import os
 import re
 import warnings
-from collections import namedtuple
-from collections.abc import Container, Iterable, Sequence, Sized
+from collections.abc import Container, Iterable, Sized
 from contextlib import contextmanager
 from functools import wraps
 from pathlib import Path
 from types import MappingProxyType
 
-# do not use yarl.quote directly,
-# use `URL(path).raw_path` instead of `quote(path)`
-# Escaping of the URLs need to be consitent with the escaping done by yarl
-from yarl import URL, unquote
+from yarl import URL
 
-from . import hdrs, helpers
+from . import hdrs
 from .abc import AbstractMatchInfo, AbstractRouter, AbstractView
+from .helpers import DEBUG
 from .http import HttpVersion11
 from .web_exceptions import (HTTPExpectationFailed, HTTPForbidden,
                              HTTPMethodNotAllowed, HTTPNotFound)
@@ -32,32 +29,13 @@ from .web_response import Response
 __all__ = ('UrlDispatcher', 'UrlMappingMatchInfo',
            'AbstractResource', 'Resource', 'PlainResource', 'DynamicResource',
            'AbstractRoute', 'ResourceRoute',
-           'StaticResource', 'View', 'RouteDef', 'RouteTableDef',
-           'head', 'get', 'post', 'patch', 'put', 'delete', 'route')
+           'StaticResource', 'View')
 
 HTTP_METHOD_RE = re.compile(r"^[0-9A-Za-z!#\$%&'\*\+\-\.\^_`\|~]+$")
 ROUTE_RE = re.compile(r'(\{[_a-zA-Z][^{}]*(?:\{[^{}]*\}[^{}]*)*\})')
 PATH_SEP = re.escape('/')
 
 
-class RouteDef(namedtuple('_RouteDef', 'method, path, handler, kwargs')):
-    def __repr__(self):
-        info = []
-        for name, value in sorted(self.kwargs.items()):
-            info.append(", {}={!r}".format(name, value))
-        return ("<RouteDef {method} {path} -> {handler.__name__!r}"
-                "{info}>".format(method=self.method, path=self.path,
-                                 handler=self.handler, info=''.join(info)))
-
-    def register(self, router):
-        if self.method in hdrs.METH_ALL:
-            reg = getattr(router, 'add_'+self.method.lower())
-            reg(self.path, self.handler, **self.kwargs)
-        else:
-            router.add_route(self.method, self.path, self.handler,
-                             **self.kwargs)
-
-
 class AbstractResource(Sized, Iterable):
 
     def __init__(self, *, name=None):
@@ -67,24 +45,12 @@ class AbstractResource(Sized, Iterable):
     def name(self):
         return self._name
 
-    @abc.abstractmethod  # pragma: no branch
-    def url(self, **kwargs):
-        """Construct url for resource with additional params.
-
-        Deprecated, use url_for() instead.
-
-        """
-        warnings.warn(".url(...) is deprecated, use .url_for instead",
-                      DeprecationWarning,
-                      stacklevel=3)
-
     @abc.abstractmethod  # pragma: no branch
     def url_for(self, **kwargs):
         """Construct url for resource with additional params."""
 
-    @asyncio.coroutine
     @abc.abstractmethod  # pragma: no branch
-    def resolve(self, request):
+    async def resolve(self, request):
         """Resolve resource
 
         Return (UrlMappingMatchInfo, allowed_methods) pair."""
@@ -104,6 +70,10 @@ class AbstractResource(Sized, Iterable):
     def freeze(self):
         pass
 
+    @abc.abstractmethod
+    def raw_match(self, path):
+        """Perform a raw match against path"""
+
 
 class AbstractRoute(abc.ABC):
 
@@ -112,7 +82,7 @@ class AbstractRoute(abc.ABC):
                  resource=None):
 
         if expect_handler is None:
-            expect_handler = _defaultExpectHandler
+            expect_handler = _default_expect_handler
 
         assert asyncio.iscoroutinefunction(expect_handler), \
             'Coroutine is expected, got {!r}'.format(expect_handler)
@@ -135,11 +105,10 @@ class AbstractRoute(abc.ABC):
                           "use async ones", DeprecationWarning)
 
             @wraps(handler)
-            @asyncio.coroutine
-            def handler_wrapper(*args, **kwargs):
+            async def handler_wrapper(*args, **kwargs):
                 result = old_handler(*args, **kwargs)
                 if asyncio.iscoroutine(result):
-                    result = yield from result
+                    result = await result
                 return result
             old_handler = handler
             handler = handler_wrapper
@@ -174,20 +143,8 @@ class AbstractRoute(abc.ABC):
     def url_for(self, *args, **kwargs):
         """Construct url for route with additional params."""
 
-    @abc.abstractmethod  # pragma: no branch
-    def url(self, **kwargs):
-        """Construct url for resource with additional params.
-
-        Deprecated, use url_for() instead.
-
-        """
-        warnings.warn(".url(...) is deprecated, use .url_for instead",
-                      DeprecationWarning,
-                      stacklevel=3)
-
-    @asyncio.coroutine
-    def handle_expect_header(self, request):
-        return (yield from self._expect_handler(request))
+    async def handle_expect_header(self, request):
+        return await self._expect_handler(request)
 
 
 class UrlMappingMatchInfo(dict, AbstractMatchInfo):
@@ -235,9 +192,11 @@ class UrlMappingMatchInfo(dict, AbstractMatchInfo):
 
     @contextmanager
     def set_current_app(self, app):
-        assert app in self._apps, (
-            "Expected one of the following apps {!r}, got {!r}"
-            .format(self._apps, app))
+        if DEBUG:  # pragma: no cover
+            if app not in self._apps:
+                raise RuntimeError(
+                    "Expected one of the following apps {!r}, got {!r}"
+                    .format(self._apps, app))
         prev = self._current_app
         self._current_app = app
         try:
@@ -267,8 +226,7 @@ class MatchInfoError(UrlMappingMatchInfo):
                                                 self._exception.reason)
 
 
-@asyncio.coroutine
-def _defaultExpectHandler(request):
+async def _default_expect_handler(request):
     """Default handler for Expect header.
 
     Just send "100 Continue" to client.
@@ -277,7 +235,8 @@ def _defaultExpectHandler(request):
     expect = request.headers.get(hdrs.EXPECT)
     if request.version == HttpVersion11:
         if expect.lower() == "100-continue":
-            request.writer.write(b"HTTP/1.1 100 Continue\r\n\r\n", drain=False)
+            await request.writer.write(
+                b"HTTP/1.1 100 Continue\r\n\r\n", drain=False)
         else:
             raise HTTPExpectationFailed(text="Unknown Expect: %s" % expect)
 
@@ -291,48 +250,48 @@ class Resource(AbstractResource):
     def add_route(self, method, handler, *,
                   expect_handler=None):
 
-        for route in self._routes:
-            if route.method == method or route.method == hdrs.METH_ANY:
+        for route_obj in self._routes:
+            if route_obj.method == method or route_obj.method == hdrs.METH_ANY:
                 raise RuntimeError("Added route will never be executed, "
-                                   "method {route.method} is "
-                                   "already registered".format(route=route))
+                                   "method {route.method} is already "
+                                   "registered".format(route=route_obj))
 
-        route = ResourceRoute(method, handler, self,
-                              expect_handler=expect_handler)
-        self.register_route(route)
-        return route
+        route_obj = ResourceRoute(method, handler, self,
+                                  expect_handler=expect_handler)
+        self.register_route(route_obj)
+        return route_obj
 
     def register_route(self, route):
         assert isinstance(route, ResourceRoute), \
             'Instance of Route class is required, got {!r}'.format(route)
         self._routes.append(route)
 
-    @asyncio.coroutine
-    def resolve(self, request):
+    async def resolve(self, request):
         allowed_methods = set()
 
         match_dict = self._match(request.rel_url.raw_path)
         if match_dict is None:
             return None, allowed_methods
 
-        for route in self._routes:
-            route_method = route.method
+        for route_obj in self._routes:
+            route_method = route_obj.method
             allowed_methods.add(route_method)
 
-            if (route_method == request._method or
+            if (route_method == request.method or
                     route_method == hdrs.METH_ANY):
-                return UrlMappingMatchInfo(match_dict, route), allowed_methods
+                return (UrlMappingMatchInfo(match_dict, route_obj),
+                        allowed_methods)
         else:
             return None, allowed_methods
 
-        yield  # pragma: no cover
-
     def __len__(self):
         return len(self._routes)
 
     def __iter__(self):
         return iter(self._routes)
 
+    # TODO: implement all abstract methods
+
 
 class PlainResource(Resource):
 
@@ -358,20 +317,19 @@ class PlainResource(Resource):
         else:
             return None
 
+    def raw_match(self, path):
+        return self._path == path
+
     def get_info(self):
         return {'path': self._path}
 
-    def url(self, *, query=None):
-        super().url()
-        return str(self.url_for().with_query(query))
-
     def url_for(self):
-        return URL(self._path)
+        return URL.build(path=self._path, encoded=True)
 
     def __repr__(self):
         name = "'" + self.name + "' " if self.name is not None else ""
-        return "<PlainResource {name} {path}".format(name=name,
-                                                     path=self._path)
+        return "<PlainResource {name} {path}>".format(name=name,
+                                                      path=self._path)
 
 
 class DynamicResource(Resource):
@@ -401,7 +359,7 @@ class DynamicResource(Resource):
             if '{' in part or '}' in part:
                 raise ValueError("Invalid path '{}'['{}']".format(path, part))
 
-            path = URL(part).raw_path
+            path = URL.build(path=part).raw_path
             formatter += path
             pattern += re.escape(path)
 
@@ -427,24 +385,24 @@ class DynamicResource(Resource):
         if match is None:
             return None
         else:
-            return {key: unquote(value, unsafe='+') for key, value in
-                    match.groupdict().items()}
+            return {key: URL.build(path=value, encoded=True).path
+                    for key, value in match.groupdict().items()}
+
+    def raw_match(self, path):
+        return self._formatter == path
 
     def get_info(self):
         return {'formatter': self._formatter,
                 'pattern': self._pattern}
 
     def url_for(self, **parts):
-        url = self._formatter.format_map(parts)
-        return URL(url)
-
-    def url(self, *, parts, query=None):
-        super().url(**parts)
-        return str(self.url_for(**parts).with_query(query))
+        url = self._formatter.format_map({k: URL.build(path=v).raw_path
+                                          for k, v in parts.items()})
+        return URL.build(path=url)
 
     def __repr__(self):
         name = "'" + self.name + "' " if self.name is not None else ""
-        return ("<DynamicResource {name} {formatter}"
+        return ("<DynamicResource {name} {formatter}>"
                 .format(name=name, formatter=self._formatter))
 
 
@@ -454,7 +412,7 @@ class PrefixResource(AbstractResource):
         assert not prefix or prefix.startswith('/'), prefix
         assert prefix in ('', '/') or not prefix.endswith('/'), prefix
         super().__init__(name=name)
-        self._prefix = URL(prefix).raw_path
+        self._prefix = URL.build(path=prefix).raw_path
 
     def add_prefix(self, prefix):
         assert prefix.startswith('/')
@@ -462,6 +420,11 @@ class PrefixResource(AbstractResource):
         assert len(prefix) > 1
         self._prefix = prefix + self._prefix
 
+    def raw_match(self, prefix):
+        return False
+
+    # TODO: impl missing abstract methods
+
 
 class StaticResource(PrefixResource):
     VERSION_KEY = 'v'
@@ -494,12 +457,6 @@ class StaticResource(PrefixResource):
                         'HEAD': ResourceRoute('HEAD', self._handle, self,
                                               expect_handler=expect_handler)}
 
-    def url(self, *, filename, append_version=None, query=None):
-        url = self.url_for(filename=filename, append_version=append_version)
-        if query is not None:
-            return str(url.update_query(query))
-        return str(url)
-
     def url_for(self, *, filename, append_version=None):
         if append_version is None:
             append_version = self._append_version
@@ -508,8 +465,10 @@ class StaticResource(PrefixResource):
         while filename.startswith('/'):
             filename = filename[1:]
         filename = '/' + filename
-        url = self._prefix + URL(filename).raw_path
-        url = URL(url)
+
+        # filename is not encoded
+        url = URL.build(path=self._prefix + filename)
+
         if append_version is True:
             try:
                 if filename.startswith('/'):
@@ -531,7 +490,8 @@ class StaticResource(PrefixResource):
                 return url
         return url
 
-    def _get_file_hash(self, byte_array):
+    @staticmethod
+    def _get_file_hash(byte_array):
         m = hashlib.sha256()  # todo sha256 can be configurable param
         m.update(byte_array)
         b64 = base64.urlsafe_b64encode(m.digest())
@@ -548,10 +508,9 @@ class StaticResource(PrefixResource):
             'OPTIONS', handler, self,
             expect_handler=self._expect_handler)
 
-    @asyncio.coroutine
-    def resolve(self, request):
+    async def resolve(self, request):
         path = request.rel_url.raw_path
-        method = request._method
+        method = request.method
         allowed_methods = set(self._routes)
         if not path.startswith(self._prefix):
             return None, set()
@@ -559,10 +518,10 @@ class StaticResource(PrefixResource):
         if method not in allowed_methods:
             return None, allowed_methods
 
-        match_dict = {'filename': unquote(path[len(self._prefix)+1:])}
+        match_dict = {'filename': URL.build(path=path[len(self._prefix)+1:],
+                                            encoded=True).path}
         return (UrlMappingMatchInfo(match_dict, self._routes[method]),
                 allowed_methods)
-        yield  # pragma: no cover
 
     def __len__(self):
         return len(self._routes)
@@ -570,16 +529,23 @@ class StaticResource(PrefixResource):
     def __iter__(self):
         return iter(self._routes.values())
 
-    @asyncio.coroutine
-    def _handle(self, request):
-        filename = unquote(request.match_info['filename'])
+    async def _handle(self, request):
+        rel_url = request.match_info['filename']
         try:
+            filename = Path(rel_url)
+            if filename.anchor:
+                # rel_url is an absolute name like
+                # /static/\\machine_name\c$ or /static/D:\path
+                # where the static dir is totally different
+                raise HTTPForbidden()
             filepath = self._directory.joinpath(filename).resolve()
             if not self._follow_symlinks:
                 filepath.relative_to(self._directory)
         except (ValueError, FileNotFoundError) as error:
             # relatively safe
             raise HTTPNotFound() from error
+        except HTTPForbidden:
+            raise
         except Exception as error:
             # perm error or other kind!
             request.app.logger.exception(error)
@@ -603,13 +569,13 @@ class StaticResource(PrefixResource):
         return ret
 
     def _directory_as_html(self, filepath):
-        "returns directory's index as html"
+        # returns directory's index as html
+
         # sanity check
         assert filepath.is_dir()
 
         relative_path_to_dir = filepath.relative_to(self._directory).as_posix()
         index_of = "Index of /{}".format(relative_path_to_dir)
-        head = "<head>\n<title>{}</title>\n</head>".format(index_of)
         h1 = "<h1>{}</h1>".format(index_of)
 
         index_list = []
@@ -632,13 +598,14 @@ class StaticResource(PrefixResource):
         ul = "<ul>\n{}\n</ul>".format('\n'.join(index_list))
         body = "<body>\n{}\n{}\n</body>".format(h1, ul)
 
-        html = "<html>\n{}\n{}\n</html>".format(head, body)
+        head_str = "<head>\n<title>{}</title>\n</head>".format(index_of)
+        html = "<html>\n{}\n{}\n</html>".format(head_str, body)
 
         return html
 
     def __repr__(self):
         name = "'" + self.name + "'" if self.name is not None else ""
-        return "<StaticResource {name} {path} -> {directory!r}".format(
+        return "<StaticResource {name} {path} -> {directory!r}>".format(
             name=name, path=self._prefix, directory=self._directory)
 
 
@@ -659,26 +626,20 @@ class PrefixedSubAppResource(PrefixResource):
         raise RuntimeError(".url_for() is not supported "
                            "by sub-application root")
 
-    def url(self, **kwargs):
-        """Construct url for route with additional params."""
-        raise RuntimeError(".url() is not supported "
-                           "by sub-application root")
-
     def get_info(self):
         return {'app': self._app,
                 'prefix': self._prefix}
 
-    @asyncio.coroutine
-    def resolve(self, request):
+    async def resolve(self, request):
         if not request.url.raw_path.startswith(self._prefix):
             return None, set()
-        match_info = yield from self._app.router.resolve(request)
+        match_info = await self._app.router.resolve(request)
         match_info.add_app(self._app)
         if isinstance(match_info.http_exception, HTTPMethodNotAllowed):
             methods = match_info.http_exception.allowed_methods
         else:
             methods = set()
-        return (match_info, methods)
+        return match_info, methods
 
     def __len__(self):
         return len(self._app.router.routes())
@@ -712,11 +673,6 @@ class ResourceRoute(AbstractRoute):
         """Construct url for route with additional params."""
         return self._resource.url_for(*args, **kwargs)
 
-    def url(self, **kwargs):
-        """Construct url for route with additional params."""
-        super().url(**kwargs)
-        return self._resource.url(**kwargs)
-
     def get_info(self):
         return self._resource.get_info()
 
@@ -730,9 +686,6 @@ class SystemRoute(AbstractRoute):
     def url_for(self, *args, **kwargs):
         raise RuntimeError(".url_for() is not allowed for SystemRoute")
 
-    def url(self, *args, **kwargs):
-        raise RuntimeError(".url() is not allowed for SystemRoute")
-
     @property
     def name(self):
         return None
@@ -740,8 +693,7 @@ class SystemRoute(AbstractRoute):
     def get_info(self):
         return {'http_exception': self._http_exception}
 
-    @asyncio.coroutine
-    def _handler(self, request):
+    async def _handler(self, request):
         raise self._http_exception
 
     @property
@@ -758,19 +710,17 @@ class SystemRoute(AbstractRoute):
 
 class View(AbstractView):
 
-    @asyncio.coroutine
-    def __iter__(self):
-        if self.request._method not in hdrs.METH_ALL:
+    async def _iter(self):
+        if self.request.method not in hdrs.METH_ALL:
             self._raise_allowed_methods()
-        method = getattr(self, self.request._method.lower(), None)
+        method = getattr(self, self.request.method.lower(), None)
         if method is None:
             self._raise_allowed_methods()
-        resp = yield from method()
+        resp = await method()
         return resp
 
-    if helpers.PY_35:
-        def __await__(self):
-            return (yield from self.__iter__())
+    def __await__(self):
+        return self._iter().__await__()
 
     def _raise_allowed_methods(self):
         allowed_methods = {
@@ -798,8 +748,8 @@ class RoutesView(Sized, Iterable, Container):
     def __init__(self, resources):
         self._routes = []
         for resource in resources:
-            for route in resource:
-                self._routes.append(route)
+            for route_obj in resource:
+                self._routes.append(route_obj)
 
     def __len__(self):
         return len(self._routes)
@@ -807,8 +757,8 @@ class RoutesView(Sized, Iterable, Container):
     def __iter__(self):
         yield from self._routes
 
-    def __contains__(self, route):
-        return route in self._routes
+    def __contains__(self, route_obj):
+        return route_obj in self._routes
 
 
 class UrlDispatcher(AbstractRouter, collections.abc.Mapping):
@@ -820,13 +770,12 @@ class UrlDispatcher(AbstractRouter, collections.abc.Mapping):
         self._resources = []
         self._named_resources = {}
 
-    @asyncio.coroutine
-    def resolve(self, request):
-        method = request._method
+    async def resolve(self, request):
+        method = request.method
         allowed_methods = set()
 
         for resource in self._resources:
-            match_dict, allowed = yield from resource.resolve(request)
+            match_dict, allowed = await resource.resolve(request)
             if match_dict is not None:
                 return match_dict
             else:
@@ -887,8 +836,13 @@ class UrlDispatcher(AbstractRouter, collections.abc.Mapping):
     def add_resource(self, path, *, name=None):
         if path and not path.startswith('/'):
             raise ValueError("path should be started with / or be empty")
+        # Reuse last added resource if path and name are the same
+        if self._resources:
+            resource = self._resources[-1]
+            if resource.name == name and resource.raw_match(path):
+                return resource
         if not ('{' in path or '}' in path or ROUTE_RE.search(path)):
-            url = URL(path)
+            url = URL.build(path=path)
             resource = PlainResource(url.raw_path, name=name)
             self.register_resource(resource)
             return resource
@@ -936,14 +890,10 @@ class UrlDispatcher(AbstractRouter, collections.abc.Mapping):
         Shortcut for add_route with method GET, if allow_head is true another
         route is added allowing head requests to the same endpoint
         """
+        resource = self.add_resource(path, name=name)
         if allow_head:
-            # it name is not None append -head to avoid it conflicting with
-            # the GET route below
-            head_name = name and '{}-head'.format(name)
-            self.add_route(hdrs.METH_HEAD, path, handler,
-                           name=head_name, **kwargs)
-        return self.add_route(hdrs.METH_GET, path, handler, name=name,
-                              **kwargs)
+            resource.add_route(hdrs.METH_HEAD, handler, **kwargs)
+        return resource.add_route(hdrs.METH_GET, handler, **kwargs)
 
     def add_post(self, path, handler, **kwargs):
         """
@@ -969,6 +919,12 @@ class UrlDispatcher(AbstractRouter, collections.abc.Mapping):
         """
         return self.add_route(hdrs.METH_DELETE, path, handler, **kwargs)
 
+    def add_view(self, path, handler, **kwargs):
+        """
+        Shortcut for add_route with ANY methods for a class-based view
+        """
+        return self.add_route(hdrs.METH_ANY, path, handler, **kwargs)
+
     def freeze(self):
         super().freeze()
         for resource in self._resources:
@@ -979,80 +935,5 @@ class UrlDispatcher(AbstractRouter, collections.abc.Mapping):
 
         Parameter should be a sequence of RouteDef objects.
         """
-        # TODO: add_table maybe?
-        for route in routes:
-            route.register(self)
-
-
-def route(method, path, handler, **kwargs):
-    return RouteDef(method, path, handler, kwargs)
-
-
-def head(path, handler, **kwargs):
-    return route(hdrs.METH_HEAD, path, handler, **kwargs)
-
-
-def get(path, handler, *, name=None, allow_head=True, **kwargs):
-    return route(hdrs.METH_GET, path, handler, name=name,
-                 allow_head=allow_head, **kwargs)
-
-
-def post(path, handler, **kwargs):
-    return route(hdrs.METH_POST, path, handler, **kwargs)
-
-
-def put(path, handler, **kwargs):
-    return route(hdrs.METH_PUT, path, handler, **kwargs)
-
-
-def patch(path, handler, **kwargs):
-    return route(hdrs.METH_PATCH, path, handler, **kwargs)
-
-
-def delete(path, handler, **kwargs):
-    return route(hdrs.METH_DELETE, path, handler, **kwargs)
-
-
-class RouteTableDef(Sequence):
-    """Route definition table"""
-    def __init__(self):
-        self._items = []
-
-    def __repr__(self):
-        return "<RouteTableDef count={}>".format(len(self._items))
-
-    def __getitem__(self, index):
-        return self._items[index]
-
-    def __iter__(self):
-        return iter(self._items)
-
-    def __len__(self):
-        return len(self._items)
-
-    def __contains__(self, item):
-        return item in self._items
-
-    def route(self, method, path, **kwargs):
-        def inner(handler):
-            self._items.append(RouteDef(method, path, handler, kwargs))
-            return handler
-        return inner
-
-    def head(self, path, **kwargs):
-        return self.route(hdrs.METH_HEAD, path, **kwargs)
-
-    def get(self, path, **kwargs):
-        return self.route(hdrs.METH_GET, path, **kwargs)
-
-    def post(self, path, **kwargs):
-        return self.route(hdrs.METH_POST, path, **kwargs)
-
-    def put(self, path, **kwargs):
-        return self.route(hdrs.METH_PUT, path, **kwargs)
-
-    def patch(self, path, **kwargs):
-        return self.route(hdrs.METH_PATCH, path, **kwargs)
-
-    def delete(self, path, **kwargs):
-        return self.route(hdrs.METH_DELETE, path, **kwargs)
+        for route_obj in routes:
+            route_obj.register(self)
